/*
 *  Copyright (C) Samuel Thibault <samuel.thibault@fnac.net>
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation ; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY ; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the program ; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * main.c
 *
 * a ppp analyser for Unix
 *
 * (c) Copyright 2000-2001 Samuel Thibault
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <setjmp.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include "analyse.h"

short srcfd,destfd;
struct termios srctty,desttty;
int pid;

unsigned char *source;
unsigned char *desti;
speed_t bau=B9600;
int parity;
short dofork=1;
short filter;
short quit;
short xonxoff;
extern unsigned char *current;
int parms[2];
struct termios origstdin;

struct show_s show;

struct showparms_s {
	char *name;
	char *padd;
	char shortcut;
	short *parm;
	char *descr;
} showparms[] = {
	// 'B' is for re(B)oot

	{ "xonxoff","  ",	'X',	&show.xonxoff,
	  "show whenever a XON or a XOFF character is received" },

	{ "asyncmap"," ",	'A',	&show.asyncmap,
	  "show when the peer seems to change its asyncmap policy" },

	{ "cp","       ",	'C',	&show.cp,
	  "show CP negociation" },

	{ "auth","     ",	'P',	&show.auth,
	  "show authentication negociation" },

	{ "ip","       ",	'I',	&show.ip,
	  "show IP packets' headers" },

	{ "icmp","     ",	'M',	&show.icmp,
	  "show ICMP events" },

	{ "udp","      ",	'U',	&show.udp,
	  "show UDP packets" },

	{ "tcp","      ",	'T',	&show.tcp,
	  "show TCP streams" },

	{ "domain","   ",	'D',	&show.domain,
	  "show domain name resolution" },
	  
	{ "others","   ",	'O',	&show.others,
	  "show others things that are still not supported" },

	{ NULL, NULL,		' ',	NULL,
	  NULL}
};

char *devname (char *s) {
	char *p,*q;
	q=s;
	for (p=s;*p;p++) if (*p=='/') (q=p+1);
	return q;
}

void syntax(char *nomprog) {
	struct showparms_s *cur;
	fprintf(stderr,"\
Syntax : %s [options] [--] source [options] desti\n\
 options:\n\
\n\
Line parameters:\n\
  -b baurate   sets baudrate\n\
  -e           sets even parity\n\
  -o           sets odd parity\n\
  -x           sets xonxoff\n\
  -c           sets crtscts\n\
\n\
Forking parameters:\n\
  -f           don't fork & behave as filter (log on stderr)\n\
  -n           don't fork\n\
\n\
Filtering:\n",nomprog);
	for (cur=showparms;cur->parm;cur++)
		fprintf(stderr,"  --%s%s%s (%c)\n",cur->name,cur->padd,cur->descr,cur->shortcut);
	fprintf(stderr,"\nand source and desti are special character files, or mere labels if -f is used\n");
	exit(0);
}

static jmp_buf albuf;

static void get_alrm(int dummy) {
	(void)dummy;
	longjmp(albuf, 1);
}

/*
 * do_setup setups the tty so as to have a blocking read, but not other
 * special handling
 */

void do_setup(char *filename, short *fd, struct termios *savtty) {
	int n;
	struct termios tty;

	fprintf(stderr,"setting up %s...",filename);
	tcgetattr(*fd, &tty);
	*savtty=tty;

/* blocking */
	n = fcntl (*fd,F_GETFL,0);
	(void) fcntl(*fd, F_SETFL, n & ~O_NONBLOCK);

	tty.c_iflag = (xonxoff?IXON|IXOFF:0);
	tty.c_lflag = 0;
	tty.c_oflag = 0;
	tty.c_cflag = 0;

/* baudrate */
	cfsetospeed(&tty,(bau));
	cfsetispeed(&tty,(bau));

/* ppp requires 8-bit cleanness anyway */
	tty.c_cflag = (tty.c_cflag & ~CSIZE) |CS8;
	tty.c_cflag |= CLOCAL | CREAD | HUPCL;
	tty.c_cflag &= ~(CRTSCTS|parity);
	tcsetattr(*fd,TCSANOW, &tty);
	tcflush(*fd,TCIOFLUSH);
	fprintf(stderr," done.\r\n");
}

/*
 * setup calls do_setup, but opens an fd before, in non blocking mode
 */

void setup(char *filename, short *fd, struct termios *savtty) {
	fprintf(stderr,"opening %s...",filename);

	*fd=-1;
	if (setjmp(albuf) == 0) {
		signal(SIGALRM, get_alrm);
		alarm(2);
		*fd=open(filename,O_RDWR|O_NONBLOCK);
		if (*fd>=0) {
			fprintf(stderr," done.\r\n");
			do_setup(filename,fd,savtty);
		}
	}
	if (*fd<0) {
		perror("open");
		exit(1);
	}
	alarm(0);
	signal(SIGALRM, SIG_IGN);
}

void do_back(char *filename, short *fd, struct termios *savtty) {
	fprintf(stderr,"setting back %s...",filename);

	tcsetattr(*fd,TCSANOW, savtty);
	tcflush(*fd,TCIOFLUSH);
	fprintf(stderr," done.\r\n");
}

void back(char *filename, short *fd, struct termios *savtty) {
	do_back(filename,fd,savtty);
	fprintf(stderr,"closing %s...",filename);
	close(*fd);
	fprintf(stderr,"done.\r\n");
}

/*
 * something happened, let's quit neatly the whole thing
 */

void restore(int dummy) {
	int n;

	if (dummy) fprintf(stderr,"killed by signal %d!\r\n",dummy);
	if (dofork && pid) /* up to parent to kill child, if any */
		kill(pid,dummy);
	if (!dofork || pid) { /* up to parent to set back */
		back(source,&srcfd,&srctty);
		back(desti,&destfd,&desttty);
	}
	if (!filter) {
		n=fcntl (STDIN_FILENO,F_GETFL,0);
		(void) fcntl(STDIN_FILENO,F_SETFL, n & ~O_NONBLOCK);
		tcsetattr(STDIN_FILENO,TCSANOW,&origstdin);
	}
	exit(1);
}

/*
 * reads on stdin, if one wants to change showing parameters
 */
void readparm() {
	struct showparms_s *cur;
	char c;

	while (read(STDIN_FILENO,&c,1)==1) {
		c=c & ~('a'-'A');
		if (c=='B') {
			extern short tossing;
			extern short tossed;

			tossing=1;
			tossed=0;
		} else if (c=='H') {
			for (cur = showparms ; cur->parm ; cur++)
				fprintf(stderr,"%c: %s\n",cur->shortcut,cur->descr);
		} else {
			for (cur = showparms ; cur->parm && c != cur->shortcut ; cur++);
			if (cur->parm) {
				*cur->parm = !*cur->parm;
				if (!dofork || !pid) fprintf(stderr,"%s%s\n",*cur->parm ? "" : "don't ", cur->descr);
				write(STDOUT_FILENO,&c,1);
			}
		}
	}
}

/*
 * reads a char on input, outputs it, and gives it to the analyser
 */
void readchar() {
	int n;
	char c;

	n=read(srcfd,&c,1);
	if (n == 1) {
		n=write(destfd,&c,1);
		if (n != 1) {
			fprintf(stderr,"write error: %d,%d\r\n",errno,n);
			exit(1);
		}
		analyse(c);
	} else {
		perror("read");
		exit(1);
	}
}

int main(int argc,char *argv[]) {
	char *tmps;
	int tmpfd;
	fd_set to_look,looked;
#ifdef RUNTIME
	int src=0;
	int dest=0;
	int i;
	short options_finies=0;
	struct showparms_s *cur;

	for (i=1;i<argc;i++) {
		if (options_finies || (argv[i][0]!='-')) {
			if (src)
				if (dest)
					syntax(argv[0]);	// too many names !
				else dest = i;
			else src=i;
		} else {
			switch (argv[i][1]) {
				case '-':
					if (argv[i][2]) {
						for (cur=showparms;cur->name && strcmp(cur->name,&argv[i][2]);cur++);
						if (cur->parm)
							*cur->parm=1;
						else
							if (strcmp("all",&argv[i][2]))
								syntax(argv[0]);
							else
								for (cur=showparms+1;cur->parm;cur++)
									*cur->parm=1;
					} else
						options_finies=1;
					break;
				/*
				 * Sets the baud rate
				 */
				case 'b':
					if (i>=argc-1) syntax(argv[0]);
					i++;
					bau=atol(argv[i]) / 1000;
					if (!bau && *argv[i] !='0') syntax(argv[0]);
					switch (bau) {
						case 1:	bau=B1200 ; break;
						case 2:	bau=B2400 ; break;
						case 4:	bau=B4800 ; break;
						default:
						case 9:	bau=B9600 ; break;
						case 19:	bau=B19200 ; break;
						case 38:	bau=B38400 ; break;
						case 57: 	bau=B57600 ; break;
						case 115:	bau=B115200 ; break;
					}
					break;
				/*
				 * Some parity ?
				 */
				case 'e':
					parity=PARENB;
					break;
				case 'o':
					parity=PARENB|PARODD;
					break;
				/*
				 * We'll act as a filter
				 */
				case 'f':
					filter=1;
					dofork=0;
					break;
				/*
				 * No forking !
				 */
				case 'n':
					dofork=0;
					break;
				case 'x':
					xonxoff=1;
					break;
				default:
					syntax(argv[0]);
					break;
			}
		}
	}
	if (!dest) syntax(argv[0]);			// not enough names !

	source=argv[src];
	desti=argv[dest];
#else
	source="/dev/ttyS0";
	desti="/dev/ttyS1";
#endif /* RUNTIME */
	if (filter) {
		srcfd=STDIN_FILENO;
		destfd=STDOUT_FILENO;
		do_setup(source,&srcfd,&srctty);
		do_setup(desti,&destfd,&desttty);
	} else {
		setup(source,&srcfd,&srctty);
		setup(desti,&destfd,&desttty);
	}

#ifdef RUNTIME
	source=devname(source);
	desti=devname(desti);
#else
#ifdef _XONXOFF
	show.xonxoff=1;
#endif
#ifdef _ASYNCMAP
	show.asyncmap=1;
#endif
#ifdef _CP
	show.cp=1;
#endif
#ifdef _IP
	show.ip=1;
#endif
#ifdef _ICMP
	show.icmp=1;
#endif
#ifdef _UDP
	show.udp=1;
#endif
#ifdef _TCP
	show.tcp=1;
#endif
#ifdef _DOMAIN
	show.domain=1;
#endif
#ifdef _OTHERS
	show.others=1;
#endif
	source= "A";
	desti= "B";
#endif /* RUNTIME */

	if (dofork) {
	// dofork implies no filter !
		pipe(parms);
		if ((pid = fork())) {	/* parent */
			close(STDOUT_FILENO);
			dup(parms[1]);
		} else {		/* child */
			int fd;

			tmps=source;
			source=desti;
			desti=tmps;
			tmpfd=srcfd;
			srcfd=destfd;
			destfd=tmpfd;
			close(STDIN_FILENO);
			dup(parms[0]);
			if (!(fd=open("/dev/null",O_WRONLY))) {
				fprintf(stderr,"unable to open /dev/null !\n");
				exit(1);
			}
			close(STDOUT_FILENO);
			dup(fd);
		}
		if (pid < 0) {
			fprintf(stderr,"unable to fork !\r\n");
			exit(1);
		}
	}

	if (!filter) {
		struct termios tty;
		int n;

		n=fcntl (STDIN_FILENO,F_GETFL,0);
		(void) fcntl(STDIN_FILENO,F_SETFL, n | O_NONBLOCK);
		tcgetattr(STDIN_FILENO,&tty);
		origstdin=tty;
		tty.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ICANON);
		tcsetattr(STDIN_FILENO,TCSANOW,&tty);
	}

	fprintf(stderr,"Starting analyser from %s to %s\r\n",source,desti);
	signal(SIGTERM,restore);
	signal(SIGSEGV,restore);
	signal(SIGINT,restore);
	signal(SIGHUP,restore);
	signal(SIGQUIT,restore);
	signal(SIGPIPE,restore);
	signal(SIGUSR1,restore);
	signal(SIGUSR2,restore);
	current=malloc(maxline);

	if (!current) {
		fprintf(stderr,"error : malloc %d",errno);
		restore(SIGSEGV);
	}

	if (filter) while (1) readchar();
	else {
		FD_ZERO(&to_look);
		FD_SET(STDIN_FILENO,&to_look);
		FD_SET(srcfd,&to_look);
		while (1) {
			looked=to_look;
			if (select(srcfd+1,&looked,NULL,NULL,NULL)<0) {
				fprintf(stderr,"error select: %d",errno);
				exit(1);
			}
			if (FD_ISSET(srcfd,&looked)) readchar();
			if (FD_ISSET(STDIN_FILENO,&looked)) readparm();
		}
	}
	restore(0);
	return(0);
}
